/* schierke.c - Don Yang (uguu.org)

   10/30/11
                                                                           */

#include<time.h>
#include<stdlib.h>
#include"SDL.h"

#define A 64
#define B 256

#define Z(a, b, c) a ## b ## c

/* All particles, mapping from cycle to element to swarm of points to (x, y).
   At each frame, we update points in the next cycle, and raw particles from
   the previous cycle at a slightly dimmer color.

   Each value is a fixed point value, with higher 16 bits being the screen
   coordinates, and the lower 16 bits to be the fractional part.

   All swarm particles with point index > 0 follows the 0th point.           */
typedef Sint32 i;
i p[A][4][B][2],
       d[4][B][2],
       q = 0,

/* Targets for swarm point 0, same fixed point format as above */
       b[4][2],

/* Shared indices */
       S, C, H, I, e, R, K, E = 1,
       w = 800, h = 600,

/* Drawing state */
       G, u,
       t,
       s = SDL_HWSURFACE|SDL_DOUBLEBUF,

/* Color palette for each elemental */
       P[4][3] =
{
   {255, 16, 16},
   {0, 255, 0},
   {255, 255, 64},
   {64, 128, 255}
};

SDL_Event n;
const SDL_VideoInfo *v;
SDL_Surface *o;


/* Return a random number in the range of [x, y) */
i O(i x, i y)
{
   return (y - x) * (float)rand() / RAND_MAX + x;
}

/* Update velocity for a single point to approach a target point */
void g(i z, i x, i y)
{
   if( p[q][I][S][C] < z )
      d[I][S][C] += O(x, y);
   if( p[q][I][S][C] > z )
      d[I][S][C] -= O(x, y);
   e = d[I][S][C];
   d[I][S][C] = e < -H ? -H : (e > H ? H : e);
}

/* Render a single frame */
#define r(type) \
{                                                                            \
   /* At each cycle, we will update points such that cycle points at the     \
      latest set of points.  Since we want to draw points from oldest        \
      cycle to newest cycle, we will start at cycle+1.                   */  \
   for(S = 1; S <= A; S++)                                                   \
   {                                                                         \
      C = (q + S) % A;                                                       \
      for(I = 0; I < 4; I++)                                                 \
      {                                                                      \
         /* The leader point will have shades of white */                    \
         t = SDL_MapRGB(o->format,                                           \
                        S * 255 / A,                                         \
                        S * 255 / A,                                         \
                        S * 255 / A);                                        \
         for(H = 0; H < B; H++)                                              \
         {                                                                   \
            /* After drawing the leader point, set color to draw             \
               follower points.                                  */          \
            if( H == 1 )                                                     \
               t = SDL_MapRGB(o->format,                                     \
                              P[I][0] * S / A,                               \
                              P[I][1] * S / A,                               \
                              P[I][2] * S / A);                              \
                                                                             \
            /* Scale coordinates and clip */                                 \
            e = p[C][I][H][0];                                               \
            e >>= 16;                                                        \
            if( e >= 0 && e < w )                                            \
            {                                                                \
               R = p[C][I][H][1];                                            \
               R >>= 16;                                                     \
               if( R >= 0 && R < h )                                         \
                  /* Draw pixel */                                           \
                  *(type*)((Uint8*)(o->pixels) +                             \
                                    R * o->pitch +                           \
                                    e * K)                                   \
                     = (type)t;                                              \
            }                                                                \
         }                                                                   \
      }                                                                      \
   }                                                                         \
}

int main(int x, char **y)
{
   /* Initialize SDL */
   if( SDL_Init(SDL_INIT_VIDEO) < 0 )
      printf("Unable to init SDL: %s\n", SDL_GetError());
   else
   {
      atexit(SDL_Quit);

      /* Initialize surface */
      v = SDL_GetVideoInfo();
      K = v->vfmt->BytesPerPixel;
      if( K != 2 && K != 4 )
         puts("Unsupported pixel format");
      else
      {
         if( x > 1 )
            s |= SDL_FULLSCREEN;
         o = SDL_SetVideoMode(w, h, K * 8, s);
         if( o == NULL )
            printf("Error initializing screen: %s\n", SDL_GetError());
         else
         {
            SDL_WM_SetCaption("Schierke", NULL);
            SDL_ShowCursor(SDL_DISABLE);

            /* Initialize random seed */
            srand(time(NULL));

            /* Initialize swarms */
            for(I = 0; I < 4; I++)
            {
               for(S = 0; S < B; S++)
               {
                  p[0][I][S][0] = O(0, w * 0x10000);
                  for(p[C = 0][I][S][1] = O(0, h * 0x10000);
                      C < 2;
                      d[I][S][C++] = O(-0x30000, 0x30000));
               }
               for(S = 1; S < A; S++)
                  for(C = 0; C < B; C++)
                     for(H = 0; H < 2; H++)
                        p[S][I][C][H] = p[0][I][C][H];
               b[I][0] = O(0, w * 0x10000);
               b[I][1] = O(0, h * 0x10000);
            }

            for(u = SDL_GetTicks();
                E;)
            {
               u += 1000/60;

               /* Render points */
               SDL_LockSurface(o);
               SDL_FillRect(o, NULL, SDL_MapRGB(o->format, 0, 0, 0));
               if( o->format->BytesPerPixel == 4 )
                  r(Uint32)
               else
                  r(Uint16)
               SDL_UnlockSurface(o);
               SDL_Flip(o);

               /* Update animation state */
               for(I = 0; I < 4; I++)
               {
                  /* Update target for leader about once per second */
                  if( !O(0, 60) )
                  {
                     b[I][0] = O(0x1000 * w, 0xf000 * w);
                     b[I][1] = O(0x1000 * h, 0xf000 * h);
                  }
                  H = 0x20000;
                  g(b[I][S = C = 0], 0x10, 0x1000);
                  g(b[I][C = 1], 0x10, 0x1000);

                  /* Update velocities for followers */
                  H *= 2;
                  for(S = 1; S < B; S++)
                  {
                     g(p[q][I][0][C = 0], 0x20, 0x1800);
                     g(p[q][I][0][C = 1], 0x20, 0x1800);
                  }

                  /* Update point positions */
                  H = (q + 1) % A;
                  for(S = 0; S < B; S++)
                     for(C = 0; C < 2; C++)
                        p[H][I][S][C] =
                           p[q][I][S][C] +
                           d[I][S][C];
               }
               for (q = H;
                    /* Check for keyboard events */
                    SDL_PollEvent(&n);)
                  if( n.type == SDL_KEYDOWN || n.type == SDL_QUIT )
                     E = 0;

               /* Sleep until next frame */
               G = SDL_GetTicks();
               if( G < u )
                  SDL_Delay(u - G);
            }

            /* Cleanup */
            SDL_FreeSurface(o);
         }
      }
      SDL_Quit();
   }
   return 0;
}
